@@ -31,10 +31,12 @@
 #include	"md_u.h"
 #include	"md_p.h"
 
-int Create(char *mddev, int mdfd,
+int Create(struct supertype *st, char *mddev, int mdfd,
 	   int chunk, int level, int layout, unsigned long size, int raiddisks, int sparedisks,
+	   char *name,
 	   int subdevs, mddev_dev_t devlist,
-	   int runstop, int verbose, int force)
+	   int runstop, int verbose, int force,
+	   char *bitmap_file, int bitmap_chunk, int write_behind, int delay)
 {
 	/*
 	 * Create a new raid array.
@@ -49,7 +51,7 @@ int Create(char *mddev, int mdfd,
 	 * abort.
 	 *
 	 * SET_ARRAY_INFO and ADD_NEW_DISK, and
-	 * if runstop==run, or raiddisks diskswere used,
+	 * if runstop==run, or raiddisks disks were used,
 	 * RUN_ARRAY
 	 */
 	unsigned long long minsize=0, maxsize=0;
@@ -59,15 +61,19 @@ int Create(char *mddev, int mdfd,
 	mddev_dev_t dv;
 	int fail=0, warn=0;
 	struct stat stb;
-	int first_missing = MD_SB_DISKS*2;
+	int first_missing = subdevs * 2;
 	int missing_disks = 0;
-	int insert_point = MD_SB_DISKS*2; /* where to insert a missing drive */
-	mddev_dev_t moved_disk = NULL; /* the disk that was moved out of the insert point */
+	int insert_point = subdevs * 2; /* where to insert a missing drive */
+	void *super;
+	int pass;
+	int vers;
+	int rv;
+	int bitmap_fd;
 
 	mdu_array_info_t array;
 	
-
-	if (md_get_version(mdfd) < 9000) {
+	vers = md_get_version(mdfd);
+	if (vers < 9000) {
 		fprintf(stderr, Name ": Create requires md driver version 0.90.0 or later\n");
 		return 1;
 	}
@@ -96,12 +102,6 @@ int Create(char *mddev, int mdfd,
 			Name ": at least 2 raid-devices needed for level 4 or 5\n");
 		return 1;
 	}
-	if (raiddisks+sparedisks > MD_SB_DISKS) {
-		fprintf(stderr,
-			Name ": too many devices requested: %d+%d > %d\n",
-			raiddisks, sparedisks, MD_SB_DISKS);
-		return 1;
-	}
 	if (subdevs > raiddisks+sparedisks) {
 		fprintf(stderr, Name ": You have listed more devices (%d) than are in the array(%d)!\n", subdevs, raiddisks+sparedisks);
 		return 1;
@@ -119,21 +119,21 @@ int Create(char *mddev, int mdfd,
 			break;
 		case 10:
 			layout = 0x102; /* near=2, far=1 */
-			if (verbose)
+			if (verbose > 0)
 				fprintf(stderr,
 					Name ": layout defaults to n1\n");
 			break;
 		case 5:
 		case 6:
 			layout = map_name(r5layout, "default");
-			if (verbose)
+			if (verbose > 0)
 				fprintf(stderr,
 					Name ": layout defaults to %s\n", map_num(r5layout, layout));
 			break;
 		case LEVEL_FAULTY:
 			layout = map_name(faultylayout, "default");
 
-			if (verbose)
+			if (verbose > 0)
 				fprintf(stderr,
 					Name ": layout defaults to %s\n", map_num(faultylayout, layout));
 			break;
@@ -156,14 +156,14 @@ int Create(char *mddev, int mdfd,
 	case -1: /* linear */
 		if (chunk == 0) {
 			chunk = 64;
-			if (verbose)
+			if (verbose > 0)
 				fprintf(stderr, Name ": chunk size defaults to 64K\n");
 		}
 		break;
 	default: /* raid1, multipath */
 		if (chunk) {
 			chunk = 0;
-			if (verbose)
+			if (verbose > 0)
 				fprintf(stderr, Name ": chunk size ignored for this level\n");
 		}
 		break;
@@ -187,7 +187,7 @@ int Create(char *mddev, int mdfd,
 		array.working_disks++;
 		if (dnum < raiddisks)
 			array.active_disks++;
-		fd = open(dname, O_RDONLY, 0);
+		fd = open(dname, O_RDONLY|O_EXCL, 0);
 		if (fd <0 ) {
 			fprintf(stderr, Name ": Cannot open %s: %s\n",
 				dname, strerror(errno));
@@ -208,17 +208,18 @@ int Create(char *mddev, int mdfd,
 		}
 		else {
 			ldsize = dsize;
-			dsize <<= 9;
+			ldsize <<= 9;
 		}
-		if (ldsize < MD_RESERVED_SECTORS*2LL*512LL) {
+		freesize = st->ss->avail_size(st, ldsize >> 9);
+		if (freesize == 0) {
 			fprintf(stderr, Name ": %s is too small: %luK\n",
 				dname, (unsigned long)(ldsize>>10));
 			fail = 1;
 			close(fd);
 			continue;
 		}
-		freesize = MD_NEW_SIZE_SECTORS((ldsize>>9));
-		freesize /= 2;
+
+		freesize /= 2; /* convert to K */
 
 		if (size && freesize < size) {
 			fprintf(stderr, Name ": %s is smaller that given size."
@@ -235,9 +236,11 @@ int Create(char *mddev, int mdfd,
 			mindisc = dname;
 			minsize = freesize;
 		}
-		warn |= check_ext2(fd, dname);
-		warn |= check_reiser(fd, dname);
-		warn |= check_raid(fd, dname);
+		if (runstop != 1 || verbose >= 0) {
+			warn |= check_ext2(fd, dname);
+			warn |= check_reiser(fd, dname);
+			warn |= check_raid(fd, dname);
+		}
 		close(fd);
 	}
 	if (fail) {
@@ -249,20 +252,21 @@ int Create(char *mddev, int mdfd,
 			fprintf(stderr, Name ": no size and no drives given - aborting create.\n");
 			return 1;
 		}
-		if (level > 0) {
+		if (level > 0 || level == LEVEL_MULTIPATH || level == LEVEL_FAULTY) {
 			/* size is meaningful */
 			if (minsize > 0x100000000ULL) {
 				fprintf(stderr, Name ": devices too large for RAID level %d\n", level);	
 				return 1;
 			}
 			size = minsize;
-			if (verbose)
+			if (verbose > 0)
 				fprintf(stderr, Name ": size set to %luK\n", size);
 		}
 	}
 	if (level > 0 && ((maxsize-size)*100 > maxsize)) {
-		fprintf(stderr, Name ": largest drive (%s) exceed size (%luK) by more than 1%%\n",
-			maxdisc, size);
+		if (runstop != 1 || verbose >= 0)
+			fprintf(stderr, Name ": largest drive (%s) exceed size (%luK) by more than 1%%\n",
+				maxdisc, size);
 		warn = 1;
 	}
 
@@ -273,7 +277,7 @@ int Create(char *mddev, int mdfd,
 				return 1;
 			}
 		} else {
-			if (verbose)
+			if (verbose > 0)
 				fprintf(stderr, Name ": creation continuing despite oddities due to --run\n");
 		}
 	}
@@ -342,61 +346,138 @@ int Create(char *mddev, int mdfd,
 	array.layout = layout;
 	array.chunk_size = chunk*1024;
 
-	if (ioctl(mdfd, SET_ARRAY_INFO, &array)) {
+
+	if (!st->ss->init_super(st, &super, &array, name))
+		return 1;
+
+	if (bitmap_file && strcmp(bitmap_file, "internal")==0) {
+		if ((vers%100) < 2) {
+			fprintf(stderr, Name ": internal bitmaps not supported by this kernel.\n");
+			return 1;
+		}
+		if (!st->ss->add_internal_bitmap(st, super, bitmap_chunk, delay, write_behind,
+						 &array.size, 1)) {
+			fprintf(stderr, Name ": Given bitmap chunk size not supported.\n");
+			return 1;
+		}
+		bitmap_file = NULL;
+	}
+
+
+
+	if ((vers % 100) >= 1) { /* can use different versions */
+		mdu_array_info_t inf;
+		memset(&inf, 0, sizeof(inf));
+		inf.major_version = st->ss->major;
+		inf.minor_version = st->minor_version;
+		rv = ioctl(mdfd, SET_ARRAY_INFO, &inf);
+	} else 
+		rv = ioctl(mdfd, SET_ARRAY_INFO, NULL);
+	if (rv) {
 		fprintf(stderr, Name ": SET_ARRAY_INFO failed for %s: %s\n",
 			mddev, strerror(errno));
 		return 1;
 	}
-	
-	for (dnum=0, dv = devlist ; dv ; dv=(dv->next)?(dv->next):moved_disk, dnum++) {
-		int fd;
-		struct stat stb;
-		mdu_disk_info_t disk;
 
-		disk.number = dnum;
-		if (dnum == insert_point) {
-			moved_disk = dv;
+	if (bitmap_file) {
+		int uuid[4];
+
+		if (bitmap_chunk == UnSet)
+			bitmap_chunk = DEFAULT_BITMAP_CHUNK;
+
+		st->ss->uuid_from_super(uuid, super);
+		if (CreateBitmap(bitmap_file, force, (char*)uuid, bitmap_chunk,
+			delay, write_behind,
+				 array.size*2ULL /* FIXME wrong for raid10 */)) {
+			return 1;
 		}
-		disk.raid_disk = disk.number;
-		if (disk.raid_disk < raiddisks)
-			disk.state = 6; /* active and in sync */
-		else
-			disk.state = 0;
-		if (dnum == insert_point ||
-		    strcasecmp(dv->devname, "missing")==0) {
-			disk.major = 0;
-			disk.minor = 0;
-			disk.state = 1; /* faulty */
-		} else {
-			fd = open(dv->devname, O_RDONLY, 0);
-			if (fd < 0) {
-				fprintf(stderr, Name ": failed to open %s after earlier success - aborting\n",
-					dv->devname);
-				return 1;
-			}
-			fstat(fd, &stb);
-			disk.major = major(stb.st_rdev);
-			disk.minor = minor(stb.st_rdev);
-			close(fd);
+		bitmap_fd = open(bitmap_file, O_RDWR);
+		if (bitmap_fd < 0) {
+			fprintf(stderr, Name ": weird: %s cannot be openned\n",
+			       bitmap_file);
+			return 1;
 		}
-		if (ioctl(mdfd, ADD_NEW_DISK, &disk)) {
-			fprintf(stderr, Name ": ADD_NEW_DISK for %s failed: %s\n",
-				dv->devname, strerror(errno));
+		if (ioctl(mdfd, SET_BITMAP_FILE, bitmap_fd) < 0) {
+			fprintf(stderr, Name ": Cannot set bitmap file for %s: %s\n",
+				mddev, strerror(errno));
 			return 1;
 		}
-		if (dv == moved_disk && dnum != insert_point) break;
 	}
 
+
+
+	for (pass=1; pass <=2 ; pass++) {
+		mddev_dev_t moved_disk = NULL; /* the disk that was moved out of the insert point */
+
+		for (dnum=0, dv = devlist ; dv ;
+		     dv=(dv->next)?(dv->next):moved_disk, dnum++) {
+			int fd;
+			struct stat stb;
+			mdu_disk_info_t disk;
+
+			disk.number = dnum;
+			if (dnum == insert_point) {
+				moved_disk = dv;
+			}
+			disk.raid_disk = disk.number;
+			if (disk.raid_disk < raiddisks)
+				disk.state = (1<<MD_DISK_ACTIVE) |
+						(1<<MD_DISK_SYNC);
+			else
+				disk.state = 0;
+			if (dv->writemostly)
+				disk.state |= (1<<MD_DISK_WRITEMOSTLY);
+
+			if (dnum == insert_point ||
+			    strcasecmp(dv->devname, "missing")==0) {
+				disk.major = 0;
+				disk.minor = 0;
+				disk.state = (1<<MD_DISK_FAULTY);
+			} else {
+				fd = open(dv->devname, O_RDONLY|O_EXCL, 0);
+				if (fd < 0) {
+					fprintf(stderr, Name ": failed to open %s after earlier success - aborting\n",
+						dv->devname);
+					return 1;
+				}
+				fstat(fd, &stb);
+				disk.major = major(stb.st_rdev);
+				disk.minor = minor(stb.st_rdev);
+				close(fd);
+			}
+			switch(pass){
+			case 1:
+				st->ss->add_to_super(super, &disk);
+				break;
+			case 2:
+				if (disk.state == 1) break;
+				st->ss->write_init_super(st, super, &disk, dv->devname);
+
+				if (ioctl(mdfd, ADD_NEW_DISK, &disk)) {
+					fprintf(stderr, Name ": ADD_NEW_DISK for %s failed: %s\n",
+						dv->devname, strerror(errno));
+					free(super);
+					return 1;
+				}
+
+				break;
+			}
+			if (dv == moved_disk && dnum != insert_point) break;
+		}
+	}
+	free(super);
+
 	/* param is not actually used */
 	if (runstop == 1 || subdevs >= raiddisks) {
 		mdu_param_t param;
 		if (ioctl(mdfd, RUN_ARRAY, &param)) {
 			fprintf(stderr, Name ": RUN_ARRAY failed: %s\n",
 				strerror(errno));
-			Manage_runstop(mddev, mdfd, -1);
+			Manage_runstop(mddev, mdfd, -1, 0);
 			return 1;
 		}
-		fprintf(stderr, Name ": array %s started.\n", mddev);
+		if (verbose >= 0)
+			fprintf(stderr, Name ": array %s started.\n", mddev);
 	} else {
 		fprintf(stderr, Name ": not starting array - not enough devices.\n");
 	}
