@@ -274,6 +274,10 @@ void arrayline(char *line)
 	mis.devname = NULL;
 	mis.spare_group = NULL;
 	mis.autof = 0;
+	mis.next = NULL;
+	mis.st = NULL;
+	mis.bitmap_fd = -1;
+	mis.name[0] = 0;
 
 	for (w=dl_next(line); w!=line; w=dl_next(w)) {
 		if (w[0] == '/') {
@@ -292,7 +296,7 @@ void arrayline(char *line)
 					fprintf(stderr, Name ": bad uuid: %s\n", w);
 			}
 		} else if (strncasecmp(w, "super-minor=", 12)==0 ) {
-			if (mis.super_minor >= 0)
+			if (mis.super_minor != UnSet)
 				fprintf(stderr, Name ": only specify super-minor once, %s ignored.\n",
 					w);
 			else {
@@ -304,6 +308,15 @@ void arrayline(char *line)
 					mis.super_minor = UnSet;
 				}
 			}
+		} else if (strncasecmp(w, "name=", 5)==0) {
+			if (mis.name[0])
+				fprintf(stderr, Name ": only specify name once, %s ignored.\n",
+					w);
+			else if (strlen(w+5) > 32)
+				fprintf(stderr, Name ": name too long, ignoring %s\n", w);
+			else
+				strcpy(mis.name, w+5);
+
 		} else if (strncasecmp(w, "devices=", 8 ) == 0 ) {
 			if (mis.devices)
 				fprintf(stderr, Name ": only specify devices once (use a comma separated list). %s ignored\n",
@@ -328,6 +341,15 @@ void arrayline(char *line)
 		} else if (strncasecmp(w, "spares=", 7) == 0 ) {
 			/* for warning if not all spares present */
 			mis.spare_disks = atoi(w+7);
+		} else if (strncasecmp(w, "metadata=", 9) == 0) {
+			/* style of metadata on the devices. */
+			int i;
+			
+			for(i=0; superlist[i] && !mis.st; i++)
+				mis.st = superlist[i]->match_metadata_desc(w+9);
+
+			if (!mis.st)
+				fprintf(stderr, Name ": metadata format %s unknown, ignored.\n", w+9);
 		} else if (strncasecmp(w, "auto=", 5) == 0 ) {
 			/* whether to create device special files as needed */
 			if (strcasecmp(w+5, "no")==0)
@@ -335,7 +357,7 @@ void arrayline(char *line)
 			else if (strcasecmp(w+5,"yes")==0 || strcasecmp(w+5,"md")==0)
 				mis.autof = -1;
 			else {
-				/* There might be digits, and maybe a hypen, at the end */
+				/* There might be digits, and maybe a hyphen, at the end */
 				char *e = w+5 + strlen(w+5);
 				int num = 4;
 				int len;
@@ -363,7 +385,7 @@ void arrayline(char *line)
 	}
 	if (mis.devname == NULL)
 		fprintf(stderr, Name ": ARRAY line with no device\n");
-	else if (mis.uuid_set == 0 && mis.devices == NULL && mis.super_minor == UnSet)
+	else if (mis.uuid_set == 0 && mis.devices == NULL && mis.super_minor == UnSet && mis.name[0] == 0)
 		fprintf(stderr, Name ": ARRAY line %s has no identity information.\n", mis.devname);
 	else {
 		mi = malloc(sizeof(*mi));
