@@ -37,6 +37,7 @@ extern __off64_t lseek64 __P ((int __fd, __off64_t __offset, int __whence));
 #include	<sys/stat.h>
 #include	<stdlib.h>
 #include	<time.h>
+#include	<sys/time.h>
 #include	<getopt.h>
 #include	<fcntl.h>
 #include	<stdio.h>
@@ -61,9 +62,21 @@ char *strncpy(char *dest, const char *src, size_t n) __THROW;
 #define BLKGETSIZE64 _IOR(0x12,114,size_t) /* return device size in bytes (u64 *arg) */
 #endif
 
+#define DEFAULT_BITMAP_CHUNK 4096
+#define DEFAULT_BITMAP_DELAY 5
+#define DEFAULT_MAX_WRITE_BEHIND 256
 
 #include	"md_u.h"
 #include	"md_p.h"
+#include	"bitmap.h"
+
+/* general information that might be extracted from a superblock */
+struct mdinfo {
+	mdu_array_info_t	array;
+	mdu_disk_info_t		disk;
+	__u64			events;
+	int			uuid[4];
+};
 
 #define Name "mdadm"
 
@@ -99,6 +112,7 @@ typedef struct mddev_ident_s {
 	
 	int	uuid_set;
 	int	uuid[4];
+	char	name[33];
 
 	unsigned int super_minor;
 
@@ -108,8 +122,10 @@ typedef struct mddev_ident_s {
 	int	level;
 	unsigned int raid_disks;
 	unsigned int spare_disks;
+	struct supertype *st;
 	int	autof;		/* 1 for normal, 2 for partitioned */
 	char	*spare_group;
+	int	bitmap_fd;
 
 	struct mddev_ident_s *next;
 } *mddev_ident_t;
@@ -120,6 +136,8 @@ typedef struct mddev_dev_s {
 	char disposition;	/* 'a' for add, 'r' for remove, 'f' for fail.
 				 * Not set for names read from .config
 				 */
+	char writemostly;
+	char re_add;
 	struct mddev_dev_s *next;
 } *mddev_dev_t;
 
@@ -154,6 +172,40 @@ extern mapping_t r5layout[], pers[], modes[], faultylayout[];
 
 extern char *map_dev(int major, int minor);
 
+
+extern struct superswitch {
+	void (*examine_super)(void *sbv);
+	void (*brief_examine_super)(void *sbv);
+	void (*detail_super)(void *sbv);
+	void (*brief_detail_super)(void *sbv);
+	void (*uuid_from_super)(int uuid[4], void *sbv);
+	void (*getinfo_super)(struct mdinfo *info, mddev_ident_t ident, void *sbv);
+	int (*update_super)(struct mdinfo *info, void *sbv, char *update, char *devname, int verbose);
+	__u64 (*event_super)(void *sbv);
+	int (*init_super)(struct supertype *st, void **sbp, mdu_array_info_t *info, char *name);
+	void (*add_to_super)(void *sbv, mdu_disk_info_t *dinfo);
+	int (*store_super)(struct supertype *st, int fd, void *sbv);
+	int (*write_init_super)(struct supertype *st, void *sbv, mdu_disk_info_t *dinfo, char *devname);
+	int (*compare_super)(void **firstp, void *secondv);
+	int (*load_super)(struct supertype *st, int fd, void **sbp, char *devname);
+	struct supertype * (*match_metadata_desc)(char *arg);
+	__u64 (*avail_size)(struct supertype *st, __u64 size);
+	int (*add_internal_bitmap)(struct supertype *st, void *sbv, int chunk, int delay, int write_behind, int *sizep, int may_change);
+	void (*locate_bitmap)(struct supertype *st, int fd, void *sbv);
+	int (*write_bitmap)(struct supertype *st, int fd, void *sbv);
+	int major;
+	int swapuuid; /* true if uuid is bigending rather than hostendian */
+} super0, super1, *superlist[];
+
+struct supertype {
+	struct superswitch *ss;
+	int minor_version;
+	int max_devs;
+};
+
+extern struct supertype *super_by_version(int vers, int minor);
+extern struct supertype *guess_super(int fd);
+
 #ifdef UCLIBC
   struct FTW {};
 # define FTW_PHYS 1
@@ -173,11 +225,12 @@ extern int Manage_runstop(char *devname, int fd, int runstop, int quiet);
 extern int Manage_resize(char *devname, int fd, long long size, int raid_disks);
 extern int Manage_reconfig(char *devname, int fd, int layout);
 extern int Manage_subdevs(char *devname, int fd,
-			  mddev_dev_t devlist);
+			  mddev_dev_t devlist, int verbose);
 extern int Grow_Add_device(char *devname, int fd, char *newdev);
+extern int Grow_addbitmap(char *devname, int fd, char *file, int chunk, int delay, int write_behind);
 
 
-extern int Assemble(char *mddev, int mdfd,
+extern int Assemble(struct supertype *st, char *mddev, int mdfd,
 		    mddev_ident_t ident,
 		    char *conffile,
 		    mddev_dev_t devlist,
@@ -187,17 +240,21 @@ extern int Assemble(char *mddev, int mdfd,
 
 extern int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 		 int raiddisks,
-		 mddev_dev_t devlist, int assume_clean);
+		 mddev_dev_t devlist, int assume_clean,
+		 char *bitmap_file, int bitmap_chunk, int write_behind, int delay, int verbose);
 
 
-extern int Create(char *mddev, int mdfd,
+extern int Create(struct supertype *st, char *mddev, int mdfd,
 		  int chunk, int level, int layout, unsigned long size, int raiddisks, int sparedisks,
+		  char *name,
 		  int subdevs, mddev_dev_t devlist,
-		  int runstop, int verbose, int force);
+		  int runstop, int verbose, int force,
+		  char *bitmap_file, int bitmap_chunk, int write_behind, int delay);
 
 extern int Detail(char *dev, int brief, int test);
 extern int Query(char *dev);
-extern int Examine(mddev_dev_t devlist, int brief, int scan, int SparcAdjust);
+extern int Examine(mddev_dev_t devlist, int brief, int scan, int SparcAdjust,
+		   struct supertype *forcest);
 extern int Monitor(mddev_dev_t devlist,
 		   char *mailaddr, char *alert_cmd,
 		   int period, int daemonise, int scan, int oneshot,
@@ -205,6 +262,12 @@ extern int Monitor(mddev_dev_t devlist,
 
 extern int Kill(char *dev, int force);
 
+extern int CreateBitmap(char *filename, int force, char uuid[16],
+			unsigned long chunksize, unsigned long daemon_sleep,
+			unsigned long write_behind,
+			unsigned long long array_size);
+extern int ExamineBitmap(char *filename, int brief, struct supertype *st);
+
 extern int md_get_version(int fd);
 extern int get_linux_version(void);
 extern int parse_uuid(char *str, int uuid[4]);
@@ -224,13 +287,12 @@ extern char *conf_line(FILE *file);
 extern char *conf_word(FILE *file, int allow_key);
 extern void free_line(char *line);
 extern int match_oneof(char *devices, char *devname);
-extern int load_super(int fd, mdp_super_t *super);
 extern void uuid_from_super(int uuid[4], mdp_super_t *super);
-extern int same_uuid(int a[4], int b[4]);
-extern int compare_super(mdp_super_t *first, mdp_super_t *second);
-extern unsigned long calc_sb_csum(mdp_super_t *super);
-extern int store_super(int fd, mdp_super_t *super);
-extern int enough(int level, int raid_disks, int avail_disks);
+extern int same_uuid(int a[4], int b[4], int swapuuid);
+/* extern int compare_super(mdp_super_t *first, mdp_super_t *second);*/
+extern unsigned long calc_csum(void *super, int bytes);
+extern int enough(int level, int raid_disks, int layout,
+		   char *avail, int avail_disks);
 extern int ask(char *mesg);
 
 
