@@ -28,6 +28,7 @@
  */
 
 #include "mdadm.h"
+#include <asm/byteorder.h>
 
 /*
  * All handling for the 0.90.0 version superblock is in
@@ -51,6 +52,39 @@ static unsigned long calc_sb0_csum(mdp_super_t *super)
 	return newcsum;
 }
 
+
+void super0_swap_endian(struct mdp_superblock_s *sb)
+{
+	/* as super0 superblocks are host-endian, it is sometimes
+	 * useful to be able to swap the endianness 
+	 * as (almost) everything is u32's we byte-swap every 4byte
+	 * number.
+	 * We then also have to swap the events_hi and events_lo
+	 */
+	char *sbc = (char *)sb;
+	__u32 t32;
+	int i;
+
+	for (i=0; i < MD_SB_BYTES ; i+=4) {
+		char t = sbc[i];
+		sbc[i] = sbc[i+3];
+		sbc[i+3] = t;
+		t=sbc[i+1];
+		sbc[i+1]=sbc[i+2];
+		sbc[i+2]=t;
+	}
+	t32 = sb->events_hi;
+	sb->events_hi = sb->events_lo;
+	sb->events_lo = t32;
+
+	t32 = sb->cp_events_hi;
+	sb->cp_events_hi = sb->cp_events_lo;
+	sb->cp_events_lo = t32;
+
+}	
+
+#ifndef MDASSEMBLE
+
 static void examine_super0(void *sbv)
 {
 	mdp_super_t *sb = sbv;
@@ -80,7 +114,9 @@ static void examine_super0(void *sbv)
 	atime = sb->utime;
 	printf("    Update Time : %.24s\n", ctime(&atime));
 	printf("          State : %s\n",
-	       (sb->state&(1<<MD_SB_CLEAN))?"clean":"dirty");
+	       (sb->state&(1<<MD_SB_CLEAN))?"clean":"active");
+	if (sb->state & (1<<MD_SB_BITMAP_PRESENT))
+		printf("Internal Bitmap : present\n");
 	printf(" Active Devices : %d\n", sb->active_disks);
 	printf("Working Devices : %d\n", sb->working_disks);
 	printf(" Failed Devices : %d\n", sb->failed_disks);
@@ -95,6 +131,10 @@ static void examine_super0(void *sbv)
 		c = map_num(r5layout, sb->layout);
 		printf("         Layout : %s\n", c?c:"-unknown-");
 	}
+	if (sb->level == 10) {
+		printf("         Layout : near=%d, far=%d\n",
+		       sb->layout&255, (sb->layout>>8)&255);
+	}
 	switch(sb->level) {
 	case 0:
 	case 4:
@@ -112,15 +152,19 @@ static void examine_super0(void *sbv)
 		mdp_disk_t *dp;
 		char *dv;
 		char nb[5];
+		int wonly;
 		if (d>=0) dp = &sb->disks[d];
 		else dp = &sb->this_disk;
-		sprintf(nb, "%4d", d);
+		snprintf(nb, sizeof(nb), "%4d", d);
 		printf("%4s %5d   %5d    %5d    %5d     ", d < 0 ? "this" :  nb,
 		       dp->number, dp->major, dp->minor, dp->raid_disk);
+		wonly = dp->state & (1<<MD_DISK_WRITEMOSTLY);
+		dp->state &= ~(1<<MD_DISK_WRITEMOSTLY);
 		if (dp->state & (1<<MD_DISK_FAULTY)) printf(" faulty");
 		if (dp->state & (1<<MD_DISK_ACTIVE)) printf(" active");
 		if (dp->state & (1<<MD_DISK_SYNC)) printf(" sync");
 		if (dp->state & (1<<MD_DISK_REMOVED)) printf(" removed");
+		if (wonly) printf(" write-mostly");
 		if (dp->state == 0) printf(" spare");
 		if ((dv=map_dev(dp->major, dp->minor)))
 			printf("   %s", dv);
@@ -167,7 +211,7 @@ static void brief_detail_super0(void *sbv)
 	else
 		printf("%08x", sb->set_uuid0);
 }
-
+#endif
 static void uuid_from_super0(int uuid[4], void * sbv)
 {
 	mdp_super_t *super = sbv;
@@ -183,7 +227,7 @@ static void uuid_from_super0(int uuid[4], void * sbv)
 	}
 }
 
-static void getinfo_super0(struct mdinfo *info, void *sbv)
+static void getinfo_super0(struct mdinfo *info, mddev_ident_t ident, void *sbv)
 {
 	mdp_super_t *sb = sbv;
 	int working = 0;
@@ -194,6 +238,7 @@ static void getinfo_super0(struct mdinfo *info, void *sbv)
 	info->array.patch_version = sb->patch_version;
 	info->array.raid_disks = sb->raid_disks;
 	info->array.level = sb->level;
+	info->array.layout = sb->layout;
 	info->array.md_minor = sb->md_minor;
 	info->array.ctime = sb->ctime;
 
@@ -201,11 +246,13 @@ static void getinfo_super0(struct mdinfo *info, void *sbv)
 	info->disk.major = sb->this_disk.major;
 	info->disk.minor = sb->this_disk.minor;
 	info->disk.raid_disk = sb->this_disk.raid_disk;
+	info->disk.number = sb->this_disk.number;
 
 	info->events = md_event(sb);
 
 	uuid_from_super0(info->uuid, sbv);
-	
+
+	ident->name[0] = 0;
 	/* work_disks is calculated rather than read directly */
 	for (i=0; i < MD_SB_DISKS; i++)
 		if ((sb->disks[i].state & (1<<MD_DISK_SYNC)) &&
@@ -229,12 +276,13 @@ static int update_super0(struct mdinfo *info, void *sbv, char *update, char *dev
 		memcpy(sb32+MD_SB_GENERIC_CONSTANT_WORDS+7,
 		       sb32+MD_SB_GENERIC_CONSTANT_WORDS+7+1,
 		       (MD_SB_WORDS - (MD_SB_GENERIC_CONSTANT_WORDS+7+1))*4);
-		fprintf (stderr, Name ": adjusting superblock of %s for 2.2/sparc compatability.\n",
-			 devname);
+		if (verbose >= 0)
+			fprintf (stderr, Name ": adjusting superblock of %s for 2.2/sparc compatability.\n",
+				 devname);
 	}
 	if (strcmp(update, "super-minor") ==0) {
 		sb->md_minor = info->array.md_minor;
-		if (verbose)
+		if (verbose > 0)
 			fprintf(stderr, Name ": updating superblock of %s with minor number %d\n",
 				devname, info->array.md_minor);
 	}
@@ -275,8 +323,10 @@ static int update_super0(struct mdinfo *info, void *sbv, char *update, char *dev
 	}
 	if (strcmp(update, "assemble")==0) {
 		int d = info->disk.number;
+		int wonly = sb->disks[d].state & (1<<MD_DISK_WRITEMOSTLY);
+		sb->disks[d].state &= ~(1<<MD_DISK_WRITEMOSTLY);
 		if (sb->disks[d].state != info->disk.state) {
-			sb->disks[d].state = info->disk.state;
+			sb->disks[d].state = info->disk.state & wonly;
 			rv = 1;
 		}
 	}
@@ -322,11 +372,12 @@ static __u64 event_super0(void *sbv)
 
 
 
-static int init_super0(void **sbp, mdu_array_info_t *info)
+static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info, char *ignored_name)
 {
-	mdp_super_t *sb = malloc(MD_SB_BYTES);
+	mdp_super_t *sb = malloc(MD_SB_BYTES + sizeof(bitmap_super_t));
 	int spares;
-	memset(sb, 0, MD_SB_BYTES);
+	int rfd;
+	memset(sb, 0, MD_SB_BYTES + sizeof(bitmap_super_t));
 
 	if (info->major_version == -1) {
 		/* zeroing the superblock */
@@ -340,12 +391,14 @@ static int init_super0(void **sbp, mdu_array_info_t *info)
 		return 0;
 	}
 
+	rfd = open("/dev/urandom", O_RDONLY);
 	sb->md_magic = MD_SB_MAGIC;
 	sb->major_version = 0;
 	sb->minor_version = 90;
 	sb->patch_version = 0;
 	sb->gvalid_words = 0; /* ignored */
-	sb->set_uuid0 = random();
+	if (rfd < 0 || read(rfd, &sb->set_uuid0, 4) != 4)
+		sb->set_uuid0 = random();
 	sb->ctime = time(0);
 	sb->level = info->level;
 	sb->size = info->size;
@@ -353,15 +406,20 @@ static int init_super0(void **sbp, mdu_array_info_t *info)
 	sb->raid_disks = info->raid_disks;
 	sb->md_minor = info->md_minor;
 	sb->not_persistent = 0;
-	sb->set_uuid1 = random();
-	sb->set_uuid2 = random();
-	sb->set_uuid3 = random();
+	if (rfd < 0 || read(rfd, &sb->set_uuid1, 12) != 12) {
+		sb->set_uuid1 = random();
+		sb->set_uuid2 = random();
+		sb->set_uuid3 = random();
+	}
+	if (rfd >= 0)
+		close(rfd);
 
 	sb->utime = sb->ctime;
 	sb->state = info->state;
 	sb->active_disks = info->active_disks;
 	sb->working_disks = info->working_disks;
 	sb->failed_disks = info->failed_disks;
+	sb->spare_disks = info->spare_disks;
 	sb->events_hi = 0;
 	sb->events_lo = 1;
 
@@ -385,7 +443,7 @@ static void add_to_super0(void *sbv, mdu_disk_info_t *dinfo)
 	dk->state = dinfo->state;
 }
 
-static int store_super0(int fd, void *sbv)
+static int store_super0(struct supertype *st, int fd, void *sbv)
 {
 	unsigned long size;
 	unsigned long long dsize;
@@ -422,7 +480,7 @@ static int store_super0(int fd, void *sbv)
 static int write_init_super0(struct supertype *st, void *sbv, mdu_disk_info_t *dinfo, char *devname)
 {
 	mdp_super_t *sb = sbv;
-	int fd = open(devname, O_RDWR, O_EXCL);
+	int fd = open(devname, O_RDWR|O_EXCL);
 	int rv;
 
 	if (fd < 0) {
@@ -430,9 +488,16 @@ static int write_init_super0(struct supertype *st, void *sbv, mdu_disk_info_t *d
 		return -1;
 	}
 
+	sb->disks[dinfo->number].state &= ~(1<<MD_DISK_FAULTY);
+	sb->disks[dinfo->number].state |= (1<<MD_DISK_SYNC);
+
 	sb->this_disk = sb->disks[dinfo->number];
 	sb->sb_csum = calc_sb0_csum(sb);
-	rv = store_super0(fd, sb);
+	rv = store_super0(st, fd, sb);
+
+	if (rv == 0 && (sb->state & (1<<MD_SB_BITMAP_PRESENT)))
+		rv = st->ss->write_bitmap(st, fd, sbv);
+
 	close(fd);
 	if (rv)
 		fprintf(stderr, Name ": failed to write superblock to %s\n", devname);
@@ -463,7 +528,7 @@ static int compare_super0(void **firstp, void *secondv)
 
 	uuid_from_super0(uuid1, first);
 	uuid_from_super0(uuid2, second);
-	if (!same_uuid(uuid1, uuid2))
+	if (!same_uuid(uuid1, uuid2, 0))
 		return 2;
 	if (first->major_version != second->major_version ||
 	    first->minor_version != second->minor_version ||
@@ -525,7 +590,7 @@ static int load_super0(struct supertype *st, int fd, void **sbp, char *devname)
 		return 1;
 	}
 
-	super = malloc(MD_SB_BYTES);
+	super = malloc(MD_SB_BYTES + sizeof(bitmap_super_t));
 
 	if (read(fd, super, sizeof(*super)) != MD_SB_BYTES) {
 		if (devname)
@@ -535,6 +600,9 @@ static int load_super0(struct supertype *st, int fd, void **sbp, char *devname)
 		return 1;
 	}
 
+	if (st->ss && st->minor_version == 9)
+		super0_swap_endian(super);
+
 	if (super->md_magic != MD_SB_MAGIC) {
 		if (devname)
 			fprintf(stderr, Name ": No super block found on %s (Expected magic %08x, got %08x)\n",
@@ -554,6 +622,7 @@ static int load_super0(struct supertype *st, int fd, void **sbp, char *devname)
 	if (st->ss == NULL) {
 		st->ss = &super0;
 		st->minor_version = 90;
+		st->max_devs = MD_SB_DISKS;
 	}
 
 	return 0;
@@ -566,28 +635,159 @@ static struct supertype *match_metadata_desc0(char *arg)
 
 	st->ss = &super0;
 	st->minor_version = 90;
+	st->max_devs = MD_SB_DISKS;
 	if (strcmp(arg, "0") == 0 ||
 	    strcmp(arg, "0.90") == 0 ||
 	    strcmp(arg, "default") == 0
 		)
 		return st;
 
+	st->minor_version = 9; /* flag for 'byte-swapped' */
+	if (strcmp(arg, "0.swap")==0)
+		return st;
+
 	free(st);
 	return NULL;
 }
 
-static __u64 avail_size0(__u64 devsize)
+static __u64 avail_size0(struct supertype *st, __u64 devsize)
 {
 	if (devsize < MD_RESERVED_SECTORS*2)
 		return 0ULL;
 	return MD_NEW_SIZE_SECTORS(devsize);
 }
 
+static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int delay, int write_behind, int *sizep, int may_change)
+{
+	/*
+	 * The bitmap comes immediately after the superblock and must be 60K in size
+	 * at most.  The default size is between 30K and 60K
+	 *
+	 * size is in K,  chunk is in bytes !!!
+	 */
+	unsigned long long size = *sizep;
+	unsigned long long bits;
+	unsigned long long max_bits = 60*1024*8;
+	unsigned long long min_chunk;
+	mdp_super_t *sb = sbv;
+	bitmap_super_t *bms = (bitmap_super_t*)(((char*)sb) + MD_SB_BYTES);
+
+	
+	min_chunk = 4096; /* sub-page chunks don't work yet.. */
+	bits = (size * 1024)/ min_chunk +1;
+	while (bits > max_bits) {
+		min_chunk *= 2;
+		bits = (bits+1)/2;
+	}
+	if (chunk == UnSet)
+		chunk = min_chunk;
+	else if (chunk < min_chunk)
+		return 0; /* chunk size too small */
+
+	sb->state |= (1<<MD_SB_BITMAP_PRESENT);
+
+	memset(bms, sizeof(*bms), 0);
+	bms->magic = __cpu_to_le32(BITMAP_MAGIC);
+	bms->version = __cpu_to_le32(BITMAP_MAJOR);
+	uuid_from_super0((int*)bms->uuid, sb);
+	bms->chunksize = __cpu_to_le32(chunk);
+	bms->daemon_sleep = __cpu_to_le32(delay);
+	bms->sync_size = __cpu_to_le64(size<<1);
+	bms->write_behind = __cpu_to_le32(write_behind);
+
+
+
+	return 1;
+}
+		
+
+void locate_bitmap0(struct supertype *st, int fd, void *sbv)
+{
+	unsigned long long dsize;
+	unsigned long size;
+	unsigned long long offset;
+#ifdef BLKGETSIZE64
+	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
+#endif
+	{
+		if (ioctl(fd, BLKGETSIZE, &size))
+			return;
+		else
+			dsize = ((unsigned long long)size)<<9;
+	}
+
+	if (dsize < MD_RESERVED_SECTORS*2)
+		return;
+	
+	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
+
+	offset *= 512;
+
+	offset += MD_SB_BYTES;
+
+	lseek64(fd, offset, 0);
+}
+
+int write_bitmap0(struct supertype *st, int fd, void *sbv)
+{
+	unsigned long size;
+	unsigned long long dsize;
+	unsigned long long offset;
+	mdp_super_t *sb = sbv;
+    
+	int rv = 0;
+
+	int towrite, n;
+	char buf[4096];
+
+#ifdef BLKGETSIZE64
+	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
+#endif
+	{
+		if (ioctl(fd, BLKGETSIZE, &size))
+			return 1;
+		else
+			dsize = ((unsigned long long)size)<<9;
+	}
+
+	if (dsize < MD_RESERVED_SECTORS*2)
+		return -1;
+	
+	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
+
+	offset *= 512;
+
+	if (lseek64(fd, offset + 4096, 0)< 0LL)
+		return 3;
+
+
+	write(fd, ((char*)sb)+MD_SB_BYTES, sizeof(bitmap_super_t));
+	towrite = 64*1024 - MD_SB_BYTES - sizeof(bitmap_super_t);
+	memset(buf, 0xff, sizeof(buf));
+	while (towrite > 0) {
+		n = towrite;
+		if (n > sizeof(buf)) 
+			n = sizeof(buf);
+		n = write(fd, buf, n);
+		if (n > 0)
+			towrite -= n;
+		else
+			break;
+	}
+	fsync(fd);
+	if (towrite)
+		rv = -2;
+
+	return rv;
+}
+
 struct superswitch super0 = {
+#ifndef MDASSEMBLE
 	.examine_super = examine_super0,
 	.brief_examine_super = brief_examine_super0,
 	.detail_super = detail_super0,
 	.brief_detail_super = brief_detail_super0,
+#endif
 	.uuid_from_super = uuid_from_super0,
 	.getinfo_super = getinfo_super0,
 	.update_super = update_super0,
@@ -600,5 +800,9 @@ struct superswitch super0 = {
 	.load_super = load_super0,
 	.match_metadata_desc = match_metadata_desc0,
 	.avail_size = avail_size0,
+	.add_internal_bitmap = add_internal_bitmap0,
+	.locate_bitmap = locate_bitmap0,
+	.write_bitmap = write_bitmap0,
 	.major = 0,
+	.swapuuid = 0,
 };
