@@ -118,10 +118,31 @@ int get_linux_version()
 	return (a*1000000)+(b*1000)+c;
 }
 
-int enough(int level, int raid_disks, int avail_disks)
+int enough(int level, int raid_disks, int layout,
+	   char *avail, int avail_disks)
 {
+	int copies, first;
 	switch (level) {
-	case 10: return 1; /* a lie, but it is hard to tell */
+	case 10:
+		/* This is the tricky one - we need to check
+		 * which actual disks are present.
+		 */
+		copies = (layout&255)* (layout>>8);
+		first=0;
+		do {
+			/* there must be one of the 'copies' form 'first' */
+			int n = copies;
+			int cnt=0;
+			while (n--) {
+				if (avail[first])
+					cnt++;
+				first = (first+1) % raid_disks;
+			}
+			if (cnt == 0)
+				return 0;
+
+		} while (first != 0);
+		return 1;
 
 	case -4:
 		return avail_disks>= 1;
@@ -140,151 +161,33 @@ int enough(int level, int raid_disks, int avail_disks)
 	}
 }
 
-int same_uuid(int a[4], int b[4])
+int same_uuid(int a[4], int b[4], int swapuuid)
 {
-    if (a[0]==b[0] &&
-	a[1]==b[1] &&
-	a[2]==b[2] &&
-	a[3]==b[3])
-	return 1;
-    return 0;
-}
-
-void uuid_from_super(int uuid[4], mdp_super_t *super)
-{
-    uuid[0] = super->set_uuid0;
-    if (super->minor_version >= 90) {
-	uuid[1] = super->set_uuid1;
-	uuid[2] = super->set_uuid2;
-	uuid[3] = super->set_uuid3;
-    } else {
-	uuid[1] = 0;
-	uuid[2] = 0;
-	uuid[3] = 0;
-    }
-}
-
-int compare_super(mdp_super_t *first, mdp_super_t *second)
-{
-    /*
-     * return:
-     *  0 same, or first was empty, and second was copied
-     *  1 second had wrong number
-     *  2 wrong uuid
-     *  3 wrong other info
-     */
-    int uuid1[4], uuid2[4];
-    if (second->md_magic != MD_SB_MAGIC)
-	return 1;
-    if (first-> md_magic != MD_SB_MAGIC) {
-	memcpy(first, second, sizeof(*first));
-	return 0;
-    }
-
-    uuid_from_super(uuid1, first);
-    uuid_from_super(uuid2, second);
-    if (!same_uuid(uuid1, uuid2))
-	return 2;
-    if (first->major_version != second->major_version ||
-	first->minor_version != second->minor_version ||
-	first->patch_version != second->patch_version ||
-	first->gvalid_words  != second->gvalid_words  ||
-	first->ctime         != second->ctime         ||
-	first->level         != second->level         ||
-	first->size          != second->size          ||
-	first->raid_disks    != second->raid_disks    )
-	return 3;
-
-    return 0;
-}
-
-int load_super(int fd, mdp_super_t *super)
-{
-	/* try to read in the superblock
-	 *
-	 * return
-	 *   0 - success
-	 *   1 - no block size
-	 *   2 - too small
-	 *   3 - no seek
-	 *   4 - no read
-	 *   5 - no magic
-	 *   6 - wrong major version
-	 */
-	unsigned long size;
-	unsigned long long dsize;
-	unsigned long long offset;
-    
-#ifdef BLKGETSIZE64
-	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
-#endif
-	{
-		if (ioctl(fd, BLKGETSIZE, &size))
-			return 1;
-		else {
-			dsize = size;
-			dsize <<= 9;
+	if (swapuuid) {
+		/* parse uuids are hostendian.
+		 * uuid's from some superblocks are big-ending
+		 * if there is a difference, we need to swap.. 
+		 */
+		unsigned char *ac = (unsigned char *)a;
+		unsigned char *bc = (unsigned char *)b;
+		int i;
+		for (i=0; i<16; i+= 4) {
+			if (ac[i+0] != bc[i+3] ||
+			    ac[i+1] != bc[i+2] ||
+			    ac[i+2] != bc[i+1] ||
+			    ac[i+3] != bc[i+0])
+				return 0;
 		}
-	}
-
-	if (dsize < MD_RESERVED_SECTORS*2)
-		return 2;
-	
-	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
-
-	offset *= 512;
-
-	ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */
-
-	if (lseek64(fd, offset, 0)< 0LL)
-		return 3;
-
-	if (read(fd, super, sizeof(*super)) != sizeof(*super))
-		return 4;
-
-	if (super->md_magic != MD_SB_MAGIC)
-		return 5;
-
-	if (super->major_version != 0)
-		return 6;
-	return 0;
-}
-
-int store_super(int fd, mdp_super_t *super)
-{
-	unsigned long size;
-	unsigned long long dsize;
-	
-	long long offset;
-    
-#ifdef BLKGETSIZE64
-	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
-#endif
-	{
-		if (ioctl(fd, BLKGETSIZE, &size))
+		return 1;
+	} else {
+		if (a[0]==b[0] &&
+		    a[1]==b[1] &&
+		    a[2]==b[2] &&
+		    a[3]==b[3])
 			return 1;
-		else
-			dsize = ((unsigned long long)size) << 9;
+		return 0;
 	}
-
-	if (dsize < MD_RESERVED_SECTORS*2)
-		return 2;
-	
-	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
-
-	offset *= 512;
-
-	if (lseek64(fd, offset, 0)< 0LL)
-		return 3;
-
-	if (write(fd, super, sizeof(*super)) != sizeof(*super))
-		return 4;
-
-	fsync(fd);
-	return 0;
 }
-    
-
 
 int check_ext2(int fd, char *name)
 {
@@ -343,20 +246,28 @@ int check_reiser(int fd, char *name)
 
 int check_raid(int fd, char *name)
 {
-	mdp_super_t super;
+	void *super;
+	struct mdinfo info;
+	struct mddev_ident_s ident;
 	time_t crtime;
-	if (load_super(fd, &super))
-		return 0;
+	char *level;
+	struct supertype *st = guess_super(fd);
+
+	if (!st) return 0;
+	st->ss->load_super(st, fd, &super, name);
 	/* Looks like a raid array .. */
 	fprintf(stderr, Name ": %s appears to be part of a raid array:\n",
 		name);
-	crtime = super.ctime;
-	fprintf(stderr, "    level=%d devices=%d ctime=%s",
-		super.level, super.raid_disks, ctime(&crtime));
+	st->ss->getinfo_super(&info, &ident, super);
+	free(super);
+	crtime = info.array.ctime;
+	level = map_num(pers, info.array.level);
+	if (!level) level = "-unknown-";
+	fprintf(stderr, "    level=%s devices=%d ctime=%s",
+		level, info.array.raid_disks, ctime(&crtime));
 	return 1;
 }
 
-
 int ask(char *mesg)
 {
 	char *add = "";
@@ -519,19 +430,25 @@ char *map_dev(int major, int minor)
 
 #endif
 
-unsigned long calc_sb_csum(mdp_super_t *super)
+unsigned long calc_csum(void *super, int bytes)
 {
-        unsigned int  oldcsum = super->sb_csum;
 	unsigned long long newcsum = 0;
-	unsigned long csum;
 	int i;
+	unsigned int csum;
 	unsigned int *superc = (unsigned int*) super;
-	super->sb_csum = 0;
 
-	for(i=0; i<MD_SB_BYTES/4; i++)
+	for(i=0; i<bytes/4; i++)
 		newcsum+= superc[i];
 	csum = (newcsum& 0xffffffff) + (newcsum>>32);
-	super->sb_csum = oldcsum;
+#ifdef __alpha__
+/* The in-kernel checksum calculation is always 16bit on 
+ * the alpha, though it is 32 bit on i386...
+ * I wonder what it is elsewhere... (it uses and API in
+ * a way that it shouldn't).
+ */
+	csum = (csum & 0xffff) + (csum >> 16);
+	csum = (csum & 0xffff) + (csum >> 16);
+#endif
 	return csum;
 }
 
@@ -672,6 +589,72 @@ void put_md_name(char *name)
 		unlink(name);
 }
 
+
+
+struct superswitch *superlist[] = { &super0, &super1, NULL };
+
+struct supertype *super_by_version(int vers, int minor)
+{
+	struct supertype *st = malloc(sizeof(*st));
+	if (!st) return st;
+	if (vers == 0) {
+		st->ss = &super0;
+		st->max_devs = MD_SB_DISKS;
+	}
+
+	if (vers == 1) {
+		st->ss = &super1;
+		st->max_devs = 384;
+	}
+	st->minor_version = minor;
+	return st;
+}
+
+struct supertype *guess_super(int fd)
+{
+	/* try each load_super to find the best match,
+	 * and return the best superswitch
+	 */
+	struct superswitch  *ss;
+	struct supertype *st;
+	unsigned long besttime = 0;
+	int bestsuper = -1;
+	
+	void *sbp = NULL;
+	int i;
+
+	st = malloc(sizeof(*st));
+	memset(st, 0, sizeof(*st));
+	for (i=0 ; superlist[i]; i++) {
+		int rv;
+		ss = superlist[i];
+		st->ss = NULL;
+		rv = ss->load_super(st, fd, &sbp, NULL);
+		if (rv == 0) {
+			struct mdinfo info;
+			struct mddev_ident_s ident;
+			ss->getinfo_super(&info, &ident, sbp);
+			if (bestsuper == -1 ||
+			    besttime < info.array.ctime) {
+				bestsuper = i;
+				besttime = info.array.ctime;
+			}
+			free(sbp);
+		}
+	}
+	if (bestsuper != -1) {
+		int rv;
+		st->ss = NULL;
+		rv = superlist[bestsuper]->load_super(st, fd, &sbp, NULL);
+		if (rv == 0) {
+			free(sbp);
+			return st;
+		}
+	}
+	free(st);
+	return NULL;
+}
+
 #ifdef __TINYC__
 /* tinyc doesn't optimize this check in ioctl.h out ... */
 unsigned int __invalid_size_argument_for_IOC = 0;
